// code by jph

#define   ANI_DEBUG     1
#define   ANI_MOTION    2
#define   ANI_DELETE    4       // flag to remove from world
#define   ANI_BLEND     8       // flag to remove from world

#define   ANI_DT_MAX    0.08

static char ANIMATED_BUFFER[8192];

class aniMated {
public:
  aniMated(void);
  virtual ~aniMated(void);
  void relax(void);
  void motion(void);
  virtual void scene(void) {};
  virtual void draw(void);
  void draw_std(void);
  void draw_shadow(tensor<double> pnt);
  virtual void draw_dyn(void);
  void setEllipsoid(double mas,tensor<double> len);
  void setCuboid(double mas,tensor<double> len);
  void setCylinder(double mas,tensor<double> len);
  void setTorus(double mas,double sca,int rat);
  virtual char* toString(void);
  double          setMassByVolume(double den); // TODO not final design !!
  double&         mass(void);                   // =ine(3,3)
  tensor<double>  momentum_lin(void);           // linear momentum
  tensor<double>  momentum_ang(void);           // angular momentum
  double          energy_kin(void);  
  tensor<double>  link_lhs(tensor<double> vec,tensor<double> lev);
  tensor<double>  link_rhs1(tensor<double> vec);
  tensor<double>  link_rhs(tensor<double> vec);
  tensor<double>  link_ect(tensor<double> vec);
//----------------------------------------------------------------------
  tensor<double>  pos; // 4x4 state tensor (orientation, position)
  tensor<double>  vel; // 4   velocity          WORLD
  tensor<double>  ome; // 4   angular velocity  OBJECT
  tensor<double>  acc; // 4   linear accel      OBJECT
  tensor<double>  tor; // 4   torque            OBJECT
  
  tensor<double>  fva;
  tensor<double>  fot;
  tensor<double>  ine; // 4x4 inertia tensor, initialized with eye(4)
  
  //double          density; // relevant for floating objects
  tensor<double>  col; // 4   color
  int             status;
  rapid*          rap;
  meshgl*         msh;
  shadow*         shd;
  int             collision; // TODO
  uint            shader; // TODO
//-------------------------------------------------------------------------------------------------    
  static double   dt;
  static double   gravity;
  static aniMated* humanized;
  static rapid_touch collide(aniMated* ani0,aniMated* ani1);
};

aniMated::aniMated(void) {
  pos=tensor<double>::eye(4);
  vel=tensor<double>::vector(4);
  ome=tensor<double>::vector(4);
  acc=tensor<double>::vector(4);
  tor=tensor<double>::vector(4);
  
  fva=tensor<double>::vector(4);
  fot=tensor<double>::vector(4);
  ine=tensor<double>::diag(tensor<double>::vector(4).as( 1.0, 1.0, 1.0, 1.0));
  
  //density=1;
  col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, 1.0);
  // rmv
  status=0;
  //ort=random<double>::uniform()*100;
  msh=0;
  shd=0;
  rap=0;
  collision=0;
  shader=0;
}

aniMated::~aniMated(void) {
  if (shd)
    delete shd;
  if (rap)
    delete rap;
}

void aniMated::relax(void) {
  acc=0;
  tor=0;
}

void aniMated::motion(void) { // input: acc (world), tor (object)
  if (collision) // TODO
    return;
  if (rap) {
    // vel --> pos
    pos=rap->pos;
    
    assert(0<mass());
    // friction --> acc  
    acc-=glI(pos)*vel%fva;  // for avatar
    tor-=ome%fot;           // for avatar
    // acc --> vel
    vel+=pos*acc*dt;
    ome+=ine.solve( tor - ome.x(ine*ome) )*dt; // tor (object)    
    
  } else {
    assert(0<mass());
    // friction --> acc  
    acc-=glI(pos)*vel%fva;  // for avatar
    tor-=ome%fot;           // for avatar
    // acc --> vel
    vel+=pos*acc*dt;
    ome+=ine.solve( tor - ome.x(ine*ome) )*dt; // tor (object)
    
    // vel --> pos
    pos.data[12]+=vel.data[0]*dt;
    pos.data[13]+=vel.data[1]*dt;
    pos.data[14]+=vel.data[2]*dt;
    pos=pos*glExp(ome*dt);
  }
  
  if (10<ome.norm()) // FIXME TODO use exact integration!
    ome=0;
}

void aniMated::draw_std(void) {
  assert(pos.numel==16);
  assert(col.numel==4);
  glPushMatrix();
    glMultMatrixd(pos.data);
    glColor4dv(col.data);
    
    if (shader) {
      opengl_shader_use(shader);
      draw();
      opengl_shader_use(0);
    } else
      draw();

  glPopMatrix();
}

void aniMated::draw_shadow(tensor<double> pnt) {
  if (shd) {
    glPushMatrix();
      glMultMatrixd(pos.data); 
      shd->draw(pos.solve(pnt));
    glPopMatrix();
  }
}

void aniMated::draw(void) {
  if (msh)
    msh->draw();
}

void aniMated::draw_dyn(void) { // right now: call from world coords
  double alp=0.7;
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);
    glPushMatrix();
      glColor4d( 0.8, 0.8, 0.8, alp);
      glMultMatrixd(glMatrixFromVector(momentum_lin(),pos[3]).data);
      glutSolidCylinder(0,momentum_lin().norm(),.025,10);
    glPopMatrix();
    glPushMatrix();
      glColor4d( 1.0, 1.0, 0.0, alp);
      glMultMatrixd(glMatrixFromVector(momentum_ang(),pos[3]).data);
      glutSolidCylinder(0,momentum_ang().norm(),.025,10);
    glPopMatrix();
    //glPushMatrix();
    //  glMultMatrixd(pos.data);
    //  glScaled(ine(0,0),ine(1,1),ine(2,2));
    //  glDepthFunc(GL_LEQUAL); // TODO
    //  glLineWidth(1);
    //  glutWireSphere(1.005,(ine(0,0)+ine(2,2))*15,(ine(0,0)+ine(2,2)+ine(1,1)*2)*5);
    //  glDepthFunc(GL_LESS);    
    //glPopMatrix();
  glPopAttrib();  
}

// TODO basic shape models have to go into separate class
void aniMated::setEllipsoid(double mas,tensor<double> len) {
  tensor<double> ext=tensor<double>::vector(4).as(len(0),len(1),len(2), 0.0);
  if (mas<TENSOR_EPS)
    mas=4*MATH_1_3*MATH_PI*ext(0)*ext(1)*ext(2);
  ine.diag( mas*(ext(1)*ext(1)+ext(2)*ext(2))/5.0, // ok
            mas*(ext(0)*ext(0)+ext(2)*ext(2))/5.0, // ok
            mas*(ext(0)*ext(0)+ext(1)*ext(1))/5.0, // ok
            mas);
  //rap.set(aniData::meshgl_sphere_14336.map(tensor<double>::diag(ext)));
}

void aniMated::setCuboid(double mas,tensor<double> len) {
  tensor<double> ext=tensor<double>::vector(4).as(len(0),len(1),len(2), 0.0);
  if (mas<TENSOR_EPS)
    mas=8*abs(ext(0)*ext(1)*ext(2));
  ine.diag( mas*(ext(1)*ext(1)+ext(2)*ext(2))/3.0, // ok
            mas*(ext(0)*ext(0)+ext(2)*ext(2))/3.0, // ok
            mas*(ext(0)*ext(0)+ext(1)*ext(1))/3.0, // ok
            mas);
  //rap.set(aniData::meshgl_cube_13824.map(tensor<double>::diag(ext)));
}

void aniMated::setCylinder(double mas,tensor<double> len) {
  tensor<double> ext=tensor<double>::vector(4).as(len(0),len(1),len(2), 0.0);
  if (mas<TENSOR_EPS)
    mas=MATH_PI*ext(0)*ext(1)*2*ext(2);
  ine.diag( mas*(ext(1)*ext(1)/4.0+ext(2)*ext(2)/3.0), // ok
            mas*(ext(0)*ext(0)/4.0+ext(2)*ext(2)/3.0), // ok
            mas*(ext(0)*ext(0)+ext(1)*ext(1))/4.0,     // ok
            mas);
  //rap.set(aniData::meshgl_cylinder_10408.map(tensor<double>::diag(ext)));
}

void aniMated::setTorus(double mas,double sca,int rat) {
  double R=sca;
  double r=sca/rat;
  //tensor<double> ext=tensor<double>::vector(4).as(sca,sca,sca, 0.0);
  if (mas<TENSOR_EPS)
    mas=2*MATH_PI*MATH_PI*R*r*r;       // ok
  ine.diag( mas*(r*r*5   +R*R*4)/8.0,  // ok
            mas*(r*r*5   +R*R*4)/8.0,  // ok
            mas*(r*r*0.75+R*R  ),      // ok
            mas);
  //rap.set(aniData::meshgl_torus_4_11326.map(tensor<double>::diag(ext)));
}

double aniMated::setMassByVolume(double den) {
  assert(msh);
  double vol;
  tensor<double> pnt=msh->centroid(vol);  
  assert(TENSOR_NO_DIVISION<vol*den);  
  return mass()=vol*den;
}

double& aniMated::mass(void) {
  return ine.data[15];
}

tensor<double> aniMated::momentum_lin(void) {
  return vel*mass();
}

tensor<double> aniMated::momentum_ang(void) {
  return pos*ine*ome+pos[3].x(vel)*mass();
}

double aniMated::energy_kin(void) {
  return 0.5*((vel|vel*mass())+(ine*ome|ome));
}

tensor<double> aniMated::link_lhs(tensor<double> vec,tensor<double> lev) {
  return gl3m(pos*glD(vec)*ine.solve(glD(lev)*glI(pos)));
}

tensor<double> aniMated::link_rhs1(tensor<double> vec) {
  return gl3v( pos * ( glD(ome)*ome.x(vec) + glD(vec)*ine.solve(ome.x(ine*ome) - tor ) + acc ));
}

tensor<double> aniMated::link_rhs(tensor<double> vec) {
  return gl3v( pos * ( glD(ome)*ome.x(vec) + glD(vec)*ine.solve(ome.x(ine*ome) - tor ) ));
}

tensor<double> aniMated::link_ect(tensor<double> vec) {
  return gl3m(pos*glD(vec)*ine.solve(glI(pos)));
}

rapid_touch aniMated::collide(aniMated* ani0,aniMated* ani1) {
  rapid_touch tch;
  tch.num=0;
  assert( ani0->rap && ani1->rap ); // TODO inefficient here !?
  
  tensor<double> tmp; // TODO move to aniWorld (so that done only once!)
  tmp=ani0->pos;
  tmp.data[12]+=ani0->vel.data[0]*dt;
  tmp.data[13]+=ani0->vel.data[1]*dt;
  tmp.data[14]+=ani0->vel.data[2]*dt;
  ani0->rap->pos=tmp*glExp(ani0->ome*dt);

  tmp=ani1->pos;
  tmp.data[12]+=ani1->vel.data[0]*dt;
  tmp.data[13]+=ani1->vel.data[1]*dt;
  tmp.data[14]+=ani1->vel.data[2]*dt;
  ani1->rap->pos=tmp*glExp(ani1->ome*dt);
  
  if (rapid::collide(*ani0->rap,*ani1->rap,tch)) {
    ani0->collision=ani1->collision=1;
    tch.nrm*=-1; // check if consistent with homepage
    double nom=0;
    double den=0;
    tensor<double> lev0=tch.pnt-ani0->pos[3];
    tensor<double> lev1=tch.pnt-ani1->pos[3];
    tensor<double> q0=ani0->ine.solve(ani0->pos.solve(lev0.x(tch.nrm)));
    tensor<double> q1=ani1->ine.solve(ani1->pos.solve(lev1.x(tch.nrm)));
    nom+=ani0->vel|tch.nrm;
    nom-=ani1->vel|tch.nrm;
    nom+=ani0->ine*q0|ani0->ome;
    nom-=ani1->ine*q1|ani1->ome;
    den+=1/ani0->mass();
    den+=1/ani1->mass();
    den+=ani0->ine*q0|q0;
    den+=ani1->ine*q1|q1;
    double lam=2*nom/den;
    ani0->vel-=tch.nrm*lam/ani0->mass();
    ani1->vel+=tch.nrm*lam/ani1->mass();
    ani0->ome-=q0*lam;
    ani1->ome+=q1*lam;
  }
  return tch;
}

char* aniMated::toString(void) {
  sprintf(ANIMATED_BUFFER,"aniMated  mass %10.2e  E_kin %10.2e",mass(),energy_kin());
  return ANIMATED_BUFFER;
}

double aniMated::dt=0; // nothing moves !
double aniMated::gravity=9.81;
aniMated* aniMated::humanized=0;

void aniMated_normalize(tensor<aniMated*> ani) { // ok
  int c0,c1;
  //tensor<aniMated*> ani=whos();
	//center of total mass at 0	
	double fac=0;
	tensor<double> cen(1,4);
	for (c0=0;c0<ani.numel;++c0)
	  cen+=ani(c0)->pos[3]*ani(c0)->mass();
	fac=1/cen.data[3];
	cen*=fac; // inverse of total mass
	for (c0=0;c0<ani.numel;++c0)
	  for (c1=0;c1<3;c1++)
	    ani(c0)->pos.data[12+c1]-=cen.data[c1];
	//if (stage<ANISKELETON_NORMALIZE_LINEAR)
	//  return;
	// total linear momentum = 0
	cen=0;
	for (c0=0;c0<ani.numel;++c0)
	  cen+=ani(c0)->momentum_lin();
	cen*=fac; // inverse of total mass
	for (c0=0;c0<ani.numel;++c0)
    ani(c0)->vel-=cen;
  //mlin_=momentum_lin(); // should be all zeros
  //assert(mlin_==tensor<double>(1,4));
	//if (stage<ANISKELETON_NORMALIZE_ANGULAR)
	//  return;
  // total angular momentum = 0
  //tensor<double> lhs(2,4,4);
  //for (c0=0;c0<ani.numel;++c0) 
  //  lhs+=gl4r(ani(c0)->pos*ani(c0)->ine*glI(ani(c0)->pos)-glD(ani(c0)->pos[3])*glD(ani(c0)->pos[3])*ani(c0)->mass());
  //lhs(3,3)=1;
  //rotate(lhs.solve(momentum_ang()*-1.0));
  // mang_=momentum_ang(); // should be all zeros
  //assert(mang_==tensor<double>(1,4));
  //ekin_=energy_kin();
}

struct aniAppend {
  aniMated* mas;
  aniMated* slv;
  tensor<double> trf; // 4x4 slv.pos=mas.pos*trf
};
