// code by jph
// threads by jc

#define         ANIWAVE_CONVERSION  0.3

struct aniWavePart {
  float   pos0; // birth place
  float   pos1;
  float   vel0; // velocity
  float   vel1;
  float   amp;  // amplitude
  int     ang;  // next split (is compared to opengl_time - tic) / recursion => dispersion angle
  double  tic;  // birth time  
};

struct aniWaveEdge {
  float   pos0; // birth place
  float   pos1;
  float   edg0; // points to end of edge
  float   edg1;
  float   nrm0; // normal
  float   nrm1;
};

struct aniWaveRefl {
  double       tic;
  aniWavePart* wpt;
  aniWaveEdge* wed;
};

class aniWave;

class aniWavelet {
public:
  ~aniWavelet();
  void particle(tensor<double> loc,tensor<double> dir,float amp); // 360 deg
  void particle(tensor<double> loc,float amp);
  void particle(tensor<double> loc,tensor<double> dir,float wdt,float amp);
  void particle(int ang,tensor<double> loc,tensor<double> dir,float amp); // ang==1 means 90 deg
  void scene(void);
  void draw(void);
  linked<aniWavePart*>  prt;
  linked<aniWaveRefl>   tre;  // this should be a sorted thing: like treeset !!!
  linked<aniWaveEdge*>  edg;
  aniWave*              wav;
  double                spd;
#ifdef PTHREAD_H
  pthread_t             thread;
#endif
private:
  void regard(aniWavePart* wpt);
};

void* aniWavelet_run(void* atr) {
  ((aniWavelet*)atr)->scene();
  return 0;
}

class aniWave : public aniMated {
public:
  aniWave(int num,int szx,int szy,double mxx,double mxy);
  int  blit(tensor<double> ctr);
  void interact(aniMated* ani,meshgl* msh,linked<tensor<double>>* lst);
  void edges(tensor<double> pnt);
  double submerge(tensor<double> loc); // returns water depth of coordinate pnt
  void setSpeed(double spd);
  void update(void);
  //void scene(void);
  void draw(void);
  void draw_dyn(void);
  char* toString(void) { return "aniWave"; };
  aniWavelet*           let;
  int                   num;
  double                hgt;  // offset z coordinate to draw mesh, TODO should be in mesh design
  float                 cut;  // kill prticles with absolute amplitude below this value
  tensor<float>         arr;  // holds means to init the entire mesh
  double                mxx,mxy;
  int                   szx,szy;
  int                   lcx,lcy;
  int                   numel;  // max number of particles
  int                   dsr;  // distribute
  uint                  texture_wav;
  uint                  texture_ref;
  linked<tensor<double>> lnk_loc; // TODO debug only
  linked<tensor<double>> lnk_nrm; // TODO debug only
private:
  double                spd;  // TODO needed in this class (or just in lets) !?
};

aniWave::aniWave(int num,int szx,int szy,double mxx,double mxy) {
#ifndef PTHREAD_H
  num=1;
#endif  
  this->num=num;
  this->szx=szx;
  this->szy=szy;
  this->mxx=mxx;
  this->mxy=mxy;
  hgt=0;
  dsr=0;
  cut=0.05;  
  msh=new meshgl(meshgl::surface(tensor<float>(2,szx+1,szy+1))); // tensor<double>::vector(2).as( szx+0.0, szy+0.0)
  arr=msh->arr;
  //printf("ARRAY %s\n",arr.toInfo());
  lcx=pos(0,3); // initialization important for texture coordinates
  lcy=pos(1,3);
  numel=3500;
  assert(0<num);
  let=new aniWavelet[num];
  for (int c0=0;c0<num;++c0)
    let[c0].wav=this;
  spd=0;
  setSpeed(1);
}

int aniWave::blit(tensor<double> ctr) {
  int pcx=(int)max(0,ctr(0)-szx/2);
  int pcy=(int)max(0,ctr(1)-szy/2);
  int ret=lcx!=pcx||lcy!=pcy;
  if (ret) {
    float dcx=(pcx-lcx)/(float)szx;
    float dcy=(pcy-lcy)/(float)szy;
    pos(0,3)=lcx=pcx;
    pos(1,3)=lcy=pcy;
    for (int c0=0;c0<arr.numel;c0+=arr.size[0]) { // texture
      arr.data[c0+0]+=dcx;
      arr.data[c0+1]+=dcy;
    }    
  }
  return ret;
}

#define ANIWAVE_INTERP                            \
    if (0<pfx&&pfx<szx&&0<pfy&&pfy<szy) {         \
      pix=pfx;                                    \
      piy=pfy;                                    \
      lmx=pfx-pix;                                \
      lmy=pfy-piy;                                \
      c1=pix+(szx+1)*piy; c1<<=3; c1+=7;          \
      sum=0;                                      \
      sum+=dat[c1            ]*(1-lmx)*(1-lmy);   \
      sum+=dat[c1+8          ]*(  lmx)*(1-lmy);   \
      sum+=dat[c1  +8*(szx+1)]*(1-lmx)*(  lmy);   \
      sum+=dat[c1+8+8*(szx+1)]*(  lmx)*(  lmy);

void aniWave::interact(aniMated* ani,meshgl* msh,linked<tensor<double>>* lst) { // TODO meshgl* msh just for debugging!
  if (ani->msh) {
    tensor<double> frm=ani->pos;
    frm(0,3)=frm(1,3)=0; // buoyancy is easy to compute in this frame
    meshgl cut=ani->msh->map(frm).cut(); // mesh in world coordinates z is real height!! but (x=y=0)  
    assert(!cut.ind[MESH_QUAD].numel);
    if (msh)
      *msh=cut.map(glT(ani->pos(0,3),ani->pos(1,3),0));
    double vol=0;
    tensor<double> ctr=frm.solve(cut.centroid(vol));
    if (TENSOR_NO_DIVISION<vol) {
      tensor<double> inv=ani->pos.solve(tensor<double>::eye(4));
      // BEGIN: buoyancy
      double fac=vol*aniMated::gravity/ani->mass();
      tensor<double> frc=inv[2]*fac;
      ani->acc+=frc;
      ani->tor+=ctr.x(frc);
      // END: buoyancy      
      // BEGIN: particle emission
      tensor<double> nrm;
      double itv=300*aniMated::dt;
      double smp=random<double>::uniform()*itv;
      double sum=0;
      for (int c0=0;c0<cut.ind[MESH_EDGE].numel;c0+=2) {
        int e0=cut.ind[MESH_EDGE].data[c0+0];
        int e1=cut.ind[MESH_EDGE].data[c0+1];
        tensor<double> loc=(cut.point(e1)+cut.point(e0))*0.5;
        loc(0)+=ani->pos(0,3);
        loc(1)+=ani->pos(1,3);                
        nrm=cut.normal(e1)+cut.normal(e0);
        double amp=nrm|(ani->vel+ani->pos*ani->ome.x(inv*loc))*(cut.point(e1)-cut.point(e0)).norm();
        sum+=abs(amp);
        if (smp<sum && 1) {
          int tmp=++dsr;
          tmp%=num;
          let[tmp].particle(0,loc,nrm,amp*0.5);
          smp+=itv;
        }
      }
      // END: particle emission
      // BEGIN: forces
      tensor<double> lc0; // to become world coordinates!
      tensor<double> lc1; // to become world coordinates!
      tensor<double> lc2; // to become world coordinates!
      tensor<double> loc;       // world coordinates!
      //tensor<double> dir;       // world coordinates
      tensor<double> lev;
      tensor<double> vlo=inv*ani->vel;
      tensor<double> vrl;
      tensor<double> vrp;
      tensor<double> acc(1,4);
      tensor<double> tor(1,4);
      double skd=opengl_slider("skin drag",0.1);
      for (int c0=0;c0<cut.ind[MESH_TRIA].numel;c0+=3) {
        lc0=cut.point(cut.ind[MESH_TRIA].data[c0  ]);
        lc1=cut.point(cut.ind[MESH_TRIA].data[c0+1]);
        lc2=cut.point(cut.ind[MESH_TRIA].data[c0+2]);
        loc=(lc0+lc1+lc2)*MATH_1_3;
        loc(0)+=ani->pos(0,3);
        loc(1)+=ani->pos(1,3);
        lnk_loc.add(loc);
        lev=inv*loc;        
        nrm=(lc1-lc0).x(lc2-lc0); // ok        
        double are=nrm.norm();
        nrm*=1/are; // normal has length 1        
        nrm=inv*nrm;
        vrl=vlo+ani->ome.x(lev); // ok, relative speed of triangle midpoint
        vrp=vrl-nrm*(vrl|nrm); // speed parallel to triangle
        double amp=vrl|nrm*are; // resistance by moving in direction of normal
        if (0<amp)
          frc=nrm*amp*-5e-1;
        else
          frc=nrm*amp*-2e-1;
        frc+=vrp*are*-skd;
        ani->acc+=frc;
        ani->tor+=lev.x(frc);
        lnk_nrm.add(ani->pos*frc);
      }
      // END: forces
      //char str[80]
      //sprintf("%s %f",ani->toString(),sum);
      //opengl_bulletin.add( tensor<char>::string(ani->toString())&tensor<double>::string(sum) );
      //opengl_label_add(ani->pos[3],tensor<double>::string( sum ));
    }
  }
}

// init edges by giving sequence of edges
// pnt has format 4 x n  with each column x1 y2 x2 y2
void aniWave::edges(tensor<double> pnt) {
  aniWaveEdge* wed;
  for (int c0=0;c0<pnt.size[1];++c0) {
    wed=new aniWaveEdge();
    wed->pos0=pos(0,3) +pnt(0,c0); // birth place
    wed->pos1=pos(1,3) +pnt(1,c0);
    wed->edg0=pnt(2,c0)-pnt(0,c0);
    wed->edg1=pnt(3,c0)-pnt(1,c0);
    //wed->edg0=pos(0,3)+pnt(2,c0)-wed->pos0;
    //wed->edg1=pos(1,3)+pnt(3,c0)-wed->pos1;
    float nrm=sqrt(wed->edg0*wed->edg0+wed->edg1*wed->edg1);    
    wed->nrm0=-wed->edg1/nrm; // normal
    wed->nrm1=+wed->edg0/nrm;
    for (int c1=0;c1<num;++c1)
      let[c1].edg.add(wed);
  }
}

double aniWave::submerge(tensor<double> loc) {
  tensor<double> pnt=pos.solve(loc);
  float* dat=msh->arr.data;
  int   pix,piy;
  float pfx,pfy;
  float lmx,lmy;
  float sum;
  pfx=pnt.data[0];
  pfy=pnt.data[1];
  int c1;
  ANIWAVE_INTERP
    return pnt(2)-sum;
  }
  return pnt(2)-hgt;
}

void aniWave::setSpeed(double spd_) {
  spd_=max(0.1,spd_/ANIWAVE_CONVERSION);
  if (spd!=spd_) {
    for (int c0=0;c0<num;++c0) {
      let[c0].tre.empty();
      let[c0].prt.delete_empty();
      let[c0].spd=spd_;
    }
    spd=spd_;
    //printf("setSpeed %6.3f\n",spd*ANIWAVE_CONVERSION);
  }
}

//int super_cnt=0;

void aniWave::update(void) {
  lnk_loc.empty();
  lnk_nrm.empty();
  msh->arr=arr; // reset x y z coordinates  
  
  if (num==1) // covers case if pthread not included
    let[0].scene();
  else {    
    // TODO for wave: explicitly joinable // ensure that join() works!
    //    pthread_t thread[NUM_THREADS];
    //    pthread_attr_t attr;
    //    /* Initialize and set thread detached attribute */
    //    pthread_attr_init(&attr);
    //    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    //void* ret;
#ifdef PTHREAD_H    // TODO TODO TODO TODO 
    for (int c0=0;c0<num;++c0)
      pthread_create(&let[c0].thread,0,aniWavelet_run,&let[c0]);  
    int err=0;
    for (int c0=0;c0<num;++c0)
      err|=pthread_join(let[c0].thread,0); // &ret instead of 0
    assert(!err);
#endif
  }
  if (status&ANI_DEBUG)
	  for (int c0=0;c0<num;++c0)
      opengl_bulletin.add(tensor<char>::string("aniWave particles ")&tensor<int>::string( let[c0].prt.numel )&tensor<char>::string(" reflections ")&tensor<int>::string( let[c0].tre.numel ));

  // BEGIN: extend detailed wave field to full area
  // sensure height to 0 on boundary strip
  // TODO yet optimize calculation !
  int dim=(msh->arr.size[1]-szx-1)*8;
  float tdy=-pos(1,3)/szy;
  float pdy=-pos(1,3);
  float tDy=(mxy-pos(1,3)-szy)/szy;
  float pDy=(mxy-pos(1,3));
  for (int c0=1;c0<=(szx+1)*8;c0+=8) {
    msh->arr.data[      c0]+=tdy;
    msh->arr.data[5+    c0] =pdy;
    msh->arr.data[  dim+c0]+=tDy;
    msh->arr.data[5+dim+c0] =pDy;
  }
  for (int c0=0;c0<msh->arr.size[1];c0+=szx+1) {
    msh->arr(0,c0    )+=    -pos(0,3)     /szx;
    msh->arr(5,c0    ) =    -pos(0,3);
    msh->arr(0,c0+szx)+=(mxx-pos(0,3)-szx)/szx;
    msh->arr(5,c0+szx) =(mxx-pos(0,3));
  }
  msh->updateN3F();
  //printf("here\n");
  
  // BEGIN: random patterns
  if (1) {
    float nse;
    for (int c0=0;c0<szx;++c0)
    for (int c1=0;c1<szy;++c1) {
      nse=perlin((pos(0,3)+c0)*0.2,(pos(1,3)+c1)*0.2,opengl_time.s());
      if (0.95<abs(nse)) {
        int tmp=++dsr;
        tmp%=num;
        tensor<double> loc=tensor<double>::vector(2).as(pos(0,3)+c0,pos(1,3)+c1);
        let[tmp].particle(2,loc,tensor<double>::randn(1,2),0.3*(0<nse?1:-1));
      }
    }
  }
  // END: random patterns
}

void aniWave::draw(void) {
  if (status&ANI_DEBUG) {
    glPushMatrix();
      glMultMatrixd(glI(pos).data);
      aniWaveEdge* wed;
      glNormal3d( 0.0, 0.0, 1.0);
      glColor3d( 1.0, 1.0, 1.0);
      while (let[0].edg.token(wed)) {
        glBegin(GL_QUAD_STRIP);
        glVertex3d( wed->pos0-wed->nrm0           ,wed->pos1-wed->nrm1          , 1.0);
        glVertex3d( wed->pos0-wed->nrm0+wed->edg0 ,wed->pos1-wed->nrm1+wed->edg1, 1.0);
        glVertex3d( wed->pos0                     ,wed->pos1                    , 0.5);
        glVertex3d( wed->pos0+wed->edg0           ,wed->pos1          +wed->edg1, 0.5);    
        glVertex3d( wed->pos0                     ,wed->pos1                    ,-0.5);
        glVertex3d( wed->pos0+wed->edg0           ,wed->pos1          +wed->edg1,-0.5);
        glEnd();
      }      
    glPopMatrix();
  }
  
  int fnc;
  glGetIntegerv(GL_DEPTH_FUNC,&fnc);
  glDepthFunc(GL_LEQUAL);
  
  glEnable(GL_TEXTURE_GEN_S);
  glEnable(GL_TEXTURE_GEN_T);
  glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_REFLECTION_MAP);
  glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_REFLECTION_MAP);
  msh->texture=texture_ref;
  msh->draw();
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);
  
  msh->texture=texture_wav;  
  msh->draw();
  
  glDepthFunc(fnc);	
}

void aniWave::draw_dyn(void) {
  if (status&ANI_DEBUG) {
    glPushAttrib(GL_ENABLE_BIT);
      glDisable(GL_LIGHTING); 
      glDisable(GL_DEPTH_TEST);
      glPointSize(1);
      glBegin(GL_POINTS);
      glNormal3d( 0.0, 0.0, 1.0);
      for (int c0=0;c0<num;++c0)
        let[c0].draw();
      glEnd();
      tensor<double> loc;
      tensor<double> nrm;
      glLineWidth(1);
      glBegin(GL_LINES);
      //glBegin(GL_POINTS);
      glColor3d( 1.0, 0.0, 1.0);
      int sta;
      //printf("%i\n",lnk_loc.numel);
      while (lnk_loc.token(loc)) {
        sta=lnk_nrm.token(nrm);
        assert(sta);
        glVertex3dv(loc.data);
        glVertex3dv((loc+nrm).data);
      }
      //printf("%s\n",loc.toString());
      glEnd();
    glPopAttrib();
  }
}


aniWavelet::~aniWavelet(void) {
  prt.delete_empty();
}

void aniWavelet::particle(tensor<double> loc,tensor<double> dir,float amp) { // ok
  aniWavePart* wpt=new aniWavePart();
  wpt->pos0=loc.data[0];
  wpt->pos1=loc.data[1];
  double nrm=sqrt(dir.data[0]*dir.data[0]+dir.data[1]*dir.data[1]);
  assert(TENSOR_NO_DIVISION<nrm);
  double fac=spd/nrm*ANIWAVE_CONVERSION;
  wpt->vel0=dir.data[0]*fac;
  wpt->vel1=dir.data[1]*fac;
  wpt->amp=amp;
  wpt->ang=1; // first split at t=1, causes circular pattern
  wpt->tic=opengl_time.s();
  prt.add(wpt);
}

void aniWavelet::particle(tensor<double> loc,float amp) { // ok
  particle(loc,tensor<double>::randn(1,2),amp);
}

#define ANIWAVE_FRONT  1024

void aniWavelet::particle(tensor<double> loc,tensor<double> dir,float wdt,float amp) {
  aniWavePart* wpt;
  dir(2)=dir(3)=0;
  dir=dir.normal();
  double fac=spd*ANIWAVE_CONVERSION; // for instance = 8
  double tic=opengl_time.s()-ANIWAVE_FRONT/spd+0.001;
  for (float lam=-wdt;lam<wdt;lam+=0.4) {
    wpt=new aniWavePart();
    wpt->vel0=dir.data[0]*fac;
    wpt->vel1=dir.data[1]*fac;
    wpt->pos0=loc.data[0]+lam*dir.data[1]-wpt->vel0*ANIWAVE_FRONT/spd;
    wpt->pos1=loc.data[1]-lam*dir.data[0]-wpt->vel1*ANIWAVE_FRONT/spd;
    wpt->amp=amp*cos(lam/wdt*MATH_PI_2);
    wpt->ang=ANIWAVE_FRONT*2;
    wpt->tic=tic;
    prt.add(wpt);
    regard(wpt);
  }
}

// 0 -> 180 deg
// 1 -> 90 deg
// ...
// for circular patterns use particle(tensor<double> loc,tensor<double> dir,float amp) 
// TODO simplify these functions!!!
void aniWavelet::particle(int ang,tensor<double> loc,tensor<double> dir,float amp) {
  aniWavePart* wpt;
  ang=1<<ang;
  double nrm=sqrt(dir.data[0]*dir.data[0]+dir.data[1]*dir.data[1]);
  assert(TENSOR_NO_DIVISION<nrm);
  double fac=spd/nrm*ANIWAVE_CONVERSION;
  double tic=opengl_time.s()-ang/spd+0.001;
  wpt=new aniWavePart();
  wpt->vel0=dir.data[0]*fac;
  wpt->vel1=dir.data[1]*fac;
  wpt->pos0=loc.data[0]-wpt->vel0*ang/spd;
  wpt->pos1=loc.data[1]-wpt->vel1*ang/spd;
  wpt->amp=amp;
  wpt->ang=ang<<1;
  wpt->tic=tic;
  prt.add(wpt);
  regard(wpt);
}

// compute reflection schedule for particle wpt looping all edges
void aniWavelet::regard(aniWavePart* wpt) {
  float det,lam,tau;
  aniWaveEdge* wed;
  aniWaveRefl rfl;
  rfl.tic=wpt->ang+1;  
  int beg=wpt->ang/2;
  while (edg.token(wed)) {
    det=1/(wed->edg1*wpt->vel0 - wed->edg0*wpt->vel1);
    lam=( (wed->pos0 - wpt->pos0)*wed->edg1 + (wpt->pos1 - wed->pos1)*wed->edg0 )*det;    
    tau=( (wpt->pos1 - wed->pos1)*wpt->vel0 + (wed->pos0 - wpt->pos0)*wpt->vel1 )*det;
    if (0<tau&&tau<1&&beg<lam*spd&&lam*spd<wpt->ang)
      if (lam<rfl.tic) { // has more recent collision
        rfl.tic=lam;
        rfl.wed=wed;
        rfl.wpt=wpt;
      }
  }
  if (rfl.tic<wpt->ang)
    tre.add(rfl);
}

void aniWavelet::scene(void) {  
  aniWavePart* wpt;
  aniWavePart* pre;
  double toc=opengl_time.s();
  // reflection
  float sca;
  aniWaveRefl rfl;
  while (tre.token(rfl)) // TODO we need a tree list for this!
    if (rfl.wpt->tic+rfl.tic<toc) {
      sca=0;
      sca+=(rfl.wpt->pos0-rfl.wed->pos0)*rfl.wed->nrm0;
      sca+=(rfl.wpt->pos1-rfl.wed->pos1)*rfl.wed->nrm1;
      sca*=2;
      rfl.wpt->pos0-=sca*rfl.wed->nrm0;
      rfl.wpt->pos1-=sca*rfl.wed->nrm1;      
      sca=0;
      sca+=rfl.wpt->vel0*rfl.wed->nrm0;
      sca+=rfl.wpt->vel1*rfl.wed->nrm1;
      sca*=2;
      rfl.wpt->vel0-=sca*rfl.wed->nrm0;
      rfl.wpt->vel1-=sca*rfl.wed->nrm1;      
      tre.remove();
    }
  float thr=wav->cut;
  if (wav->numel<prt.numel)
    thr=wav->cut*2;
  // subdivision
  while (prt.token(wpt))
    if (wpt->ang<(toc-wpt->tic)*spd) // lhs<rhs "*sqrt(wpt->vel0*wpt->vel0+wpt->vel1*wpt->vel1)"
      if (-thr<wpt->amp&&wpt->amp<thr) {
        delete wpt;
        prt.remove();
      } else {
        double ang=MATH_PI_2/wpt->ang;
        wpt->ang<<=1;
        double rcs=cos(ang);
        double rsn=sin(ang);
        float tmp0=wpt->vel0;
        float tmp1=wpt->vel1;
        wpt->vel0= rcs*tmp0+rsn*tmp1;
        wpt->vel1=-rsn*tmp0+rcs*tmp1;
        wpt->amp*=0.5;      
        pre=new aniWavePart();
        pre->pos0=wpt->pos0;
        pre->pos1=wpt->pos1;
        pre->vel0= rcs*tmp0-rsn*tmp1;
        pre->vel1= rsn*tmp0+rcs*tmp1;
        pre->amp=wpt->amp;
        pre->ang=wpt->ang;
        pre->tic=wpt->tic;
        prt.pre(pre);
        regard(wpt);
        regard(pre);
      }
  // particles to mesh
  float* dat=wav->msh->arr.data;
  double dlt;
  int pix,piy;
  float pfx,pfy;
  float lmx,lmy;
  float amp0,amp1,amp2,fac;
  int c0,c1,c2=0;
  double pos_x=wav->pos.data[12];
  double pos_y=wav->pos.data[13];
  int szx=wav->szx;
  int szy=wav->szy;
  while (prt.token(wpt)) {
    dlt=opengl_time.s()-wpt->tic;
    pfx=wpt->pos0+dlt*wpt->vel0-pos_x;
    pfy=wpt->pos1+dlt*wpt->vel1-pos_y;
    if (0<pfx&&pfx<szx&&0<pfy&&pfy<szy) {
      pix=pfx;
      piy=pfy;
      lmx=pfx-pix;
      lmy=pfy-piy;
      amp2=wpt->amp;
      amp0=-amp2*0.1*wpt->vel0;
      amp1=-amp2*0.1*wpt->vel1;
      //pfx=amp0*amp0+amp1*amp1;
      //if (0.8<pfx) {
      //  pfx=0.8;
      //  amp0*=pfx;
      //  amp1*=pfx;
      //}
      // bilinear interpolation
      piy*=szx+1;
      fac=(1-lmx)*(1-lmy);
      c0=pix   + piy      ; c0<<=3; c0+=4;
      dat[++c0]+=amp0*fac;
      dat[++c0]+=amp1*fac;
      dat[++c0]+=amp2*fac;
      fac=lmx *(1-lmy);
      c0=pix+1 + piy      ; c0<<=3; c0+=4;
      dat[++c0]+=amp0*fac;
      dat[++c0]+=amp1*fac;
      dat[++c0]+=amp2*fac;      
      fac=(1-lmx)*   lmy ;
      c0=pix   + piy+szx+1; c0<<=3; c0+=4;
      dat[++c0]+=amp0*fac;
      dat[++c0]+=amp1*fac;
      dat[++c0]+=amp2*fac;      
      fac=lmx *   lmy ;
      c0=pix+1 + piy+szx+1; c0<<=3; c0+=4;
      dat[++c0]+=amp0*fac;
      dat[++c0]+=amp1*fac;
      dat[++c0]+=amp2*fac;
    } else
      prt.remove();
  } 
}

void aniWavelet::draw(void) {
  if (prt.numel<6000) {
    aniWavePart* wpt;
    double dlt;
    double gre;
    while (prt.token(wpt)) {
      dlt=opengl_time.s()-wpt->tic;
      gre=0.0<wpt->amp;
      glColor3d(1.0-gre,gre,0.0);
      glVertex3d( wpt->pos0+dlt*wpt->vel0, wpt->pos1+dlt*wpt->vel1, wav->hgt);
    }
  }
}
