// code by jph

class aniWorld {
private:
  static aniWorld* instance;
  linked<aniMated*> mov;
  linked<aniMated*> all;
  linked<aniAppend*> app;
public:
  static inline aniWorld* singleton(void); 
  
  void add(aniMated* ani);
  //void pre(aniMated* ani);
  void link(aniAppend* apd);
  
  void relax(void);
  void collide(void);
  void scene(void);
  linked<aniMated*>* private_mov(void); // used for buoyancy
  void motion(void);
  void display(void);
  void shadow(opengl_light* lgt);
  int intrudes(aniMated* ani); 
  void blend(void);
  
  uint orthogonalize;
};

aniWorld* aniWorld::instance=0;

aniWorld* aniWorld::singleton(void) {
  if (!instance)
    instance=new aniWorld();
  return instance;
}

void aniWorld::add(aniMated* ani) {
  if (ani->status&ANI_MOTION)
    mov.add(ani);
  all.add(ani);
}

//void aniWorld::pre(aniMated* ani) {
//  if (ani->status&ANI_MOTION)
//    mov.pre(ani);
//  all.pre(ani);
//}

void aniWorld::link(aniAppend* apd) {
  apd->trf=tensor<double>::eye(4);
  app.add(apd);
}

void aniWorld::relax(void) {
  //timer* tmr=timer::tic("relax");
  aniMated* tok;
  linked<aniMated*> del;
  while (mov.token(tok)) {
    tok->relax(); // body: acc=tor=0
    if (tok->status&ANI_DELETE)
      mov.remove();
  }
    
  while (all.token(tok)) 
    if (tok->status&ANI_DELETE) {
      del.add(tok);
      all.remove();
    }

  while (del.token(tok)) 
    delete tok;
  //tmr->toc();
}

void aniWorld::collide(void) {
  timer* tmr=timer::tic("collide");
  aniMated* ani;
  linked<aniMated*> mvn_;
  linked<aniMated*> stc_;
  while (all.token(ani))
    if (ani->rap) {
      ani->collision=0;
      if (ani->status&ANI_MOTION)
        mvn_.add(ani);
      else
        stc_.add(ani);  
    }
  tensor<aniMated*> mvn=mvn_.toTensor();
  tensor<aniMated*> stc=mvn&stc_.toTensor();  
  //tensor<double> tmp;
  //for (int c0=0;c0<stc.numel-1;++c0) {
  //  ani=stc(c0);
  //  tmp=ani->pos;
  //  tmp.data[12]+=ani->vel.data[0]*aniMated::dt;
  //  tmp.data[13]+=ani->vel.data[1]*aniMated::dt;
  //  tmp.data[14]+=ani->vel.data[2]*aniMated::dt;
  //  ani->rap->pos=tmp*glExp(ani->ome*aniMated::dt);
  //}

  //printf("%i x %i\n",mvn.numel,stc.numel);
  for (int c0=0;c0<mvn.numel;++c0)
  for (int c1=c0+1;c1<stc.numel;++c1) {
    //printf("%s vs ",mvn(c0)->toString());
    //printf("%s\n",stc(c1)->toString());
    aniMated::collide(mvn(c0),stc(c1));
  }
  //opengl_bulletin.add( tensor<int>::string(mvn.numel)&tensor<char>::string(" vs ")&tensor<int>::string(stc.numel) );
  tmr->toc();
}

void aniWorld::scene(void) {
  timer* tmr=timer::tic("scene");
  aniMated* tok;
  while (mov.token(tok)) 
    tok->scene();
  audio::mix();
  tmr->toc();
}

linked<aniMated*>* aniWorld::private_mov(void) {
  return &mov;
}

void aniWorld::motion(void) {
  timer* tmr=timer::tic("motion");
  aniMated* tok;
  while (mov.token(tok)) 
    tok->motion(); // body: move
    
  if (!(++orthogonalize%100))
    while (mov.token(tok))
      tok->pos=glOrthogonalize(tok->pos);
      
  aniAppend* apd;
  while (app.token(apd)) 
    apd->slv->pos=apd->mas->pos*apd->trf;
  tmr->toc();
}

void aniWorld::display(void) {
  timer* tmr=timer::tic("display");
  aniMated* tok;
  while (all.token(tok)) 
    if ((tok->status&ANI_BLEND)==0)
      tok->draw_std();
  tmr->toc();
}

void aniWorld::shadow(opengl_light* lgt) {
  timer* tmr=timer::tic("shadow");
  aniMated* tok;
  while (all.token(tok))
    tok->draw_shadow(lgt->pos[3]);
  tmr->toc();
}

int aniWorld::intrudes(aniMated* ani) {
  int num=0;
  rapid_touch tch;
  aniMated* tok;
  while (all.token(tok)) {
    tch=aniMated::collide(tok,ani);
    num+=tch.num;
  }
  return num;
}

void aniWorld::blend(void) {
  aniMated* tok;
  linked<aniMated*> lst;
  while (all.token(tok))
    if (tok->status&ANI_BLEND)
      lst.add(tok);
  tensor<aniMated*> vec=lst.toTensor();
  tensor<double> dpt(1,vec.numel);
  tensor<double> res=opengl_project*opengl_camera;
  for (int c0=0;c0<vec.numel;++c0)
    dpt(c0)=(res*vec(c0)->pos[3])(2);
  tensor<int> pri;
  dpt.sort(TENSOR_SORT_CBA,pri);
  for (int c0=0;c0<vec.numel;++c0)
    vec(pri(c0))->draw_std();
}
