// code by jph

#define HASH_TOTEST   7

void perlin_bmp_test() {
  
  tensor<uchar> bmp(3,3,151,103);
  
  for (int c0=0;c0<bmp.size[1];++c0) 
  for (int c1=0;c1<bmp.size[2];++c1) {
    bmp(0,c0,c1)=(perlin(c0*0.1,c1*0.1)+1)*128;
    bmp(1,c0,c1)=(perlin(c0*0.1,0.1,c1*0.1)+1)*128;
    bmp(2,c0,c1)=(perlin(c0*0.2,c1*0.1,0.3,c0*0.1)+1)*128;
    //bmp(3,c0,c1)=(perlin(c1*0.3,c0*0.1)+1)*128;
  }
  format::bmp("perlin.bmp",bmp);
}

void perlin_test() {
  timer* tmr=timer::tic("perlin");
  
  float val,mn,mx;  
  mn=1; mx=-1;
  for (int c0=0;c0<1000000;++c0) {
    val=perlin(
      random<double>::uniform()*20+random<double>::normal()*100+c0);
    mn=min(val,mn);
    mx=max(val,mx);
  }  
  printf("%f %f\n",mn,mx);
  
  mn=1; mx=-1;
  for (int c0=0;c0<5000000;++c0) {
    val=perlin(
      random<double>::uniform()*20+random<double>::normal()*100+c0/100,
      random<double>::uniform()*10+random<double>::normal()*200-c0);
    mn=min(val,mn);
    mx=max(val,mx);
  }
  printf("%f %f\n",mn,mx);
  
  mn=1; mx=-1;
  for (int c0=0;c0<1000000;++c0) {
    val=perlin(
      random<double>::uniform()*20+random<double>::normal()*100+c0,
      random<double>::uniform()*10+random<double>::normal()*200-c0,
      random<double>::uniform()*5+random<double>::normal()*50-c0
      );
    mn=min(val,mn);
    mx=max(val,mx);
  }
  printf("%f %f\n",mn,mx);
  
  mn=1; mx=-1;
  for (int c0=0;c0<1000000;++c0) {
    val=perlin(
      random<double>::uniform()*20+random<double>::normal()*100+c0,
      random<double>::uniform()*10+random<double>::normal()*200-c0,
      random<double>::uniform()*1+random<double>::normal()*20-c0,
      random<double>::uniform()*5+random<double>::normal()*50-c0
      );
    mn=min(val,mn);
    mx=max(val,mx);
  }
  printf("%f %f\n",mn,mx);
  
  tmr->toc();  
  // 5.330
  timer::stats();

}

void tensor_variance_test() {
  tensor<double> asd=tensor<double>::vector(1).as(3.0);
  printf("%s\n",asd.toString());
  printf("%s\n\n",asd.variance().toString());
  
  tensor<double> ase=tensor<double>::vector(2).as(3.0,10.0);
  printf("%s\n",ase.toString());
  printf("%s\n\n",ase.variance().toString());
  
  tensor<double> asf=tensor<double>::vector(3).as(3.0,10.0,13.0);
  printf("%s\n",asf.toString());
  printf("%s\n\n",asf.variance().toString());
}

void tensor_operatorltlteq_test() {
  tensor<double> asd(2,4,0);
  tensor<double> ap1=tensor<double>::rand(1,4);
  tensor<double> ap2=tensor<double>::rand(1,4);
  
  asd<<=ap1;
  asd<<=ap2;
  asd<<=ap1;
  printf("%s\n",asd.toString());
}

void tensor_diff_test() {
  tensor<int> asd=tensor<int>::index(3,4,3,7);
  printf("%s\n",asd.toString());
  printf("%s\n",asd.diff(0).toString());
}

void tensor_gradient_test() {
  tensor<double> asd(2,2,5);
  asd(0,1)=1;
  asd(0,2)=20;
  asd(0,4)=1;
  asd(1,1)=1;
  asd(1,2)=20;
  asd(1,4)=2;
  asd=tensor<double>::index(3,4,3,7);
  printf("%s\n\n",asd.toString());
  printf("%s\n\n",asd.gradient(0).toString());
  
}

void tensor_interp_testIMG() {
  tensor<int> prm=tensor<int>::vector(3).as(2,0,1);
  tensor<uchar> img=format::bmp("interp.bmp").transpose(prm);
  printf("%s\n",img.toInfo());
  
  tensor<int> r=img[0].to<int>();
  tensor<int> g=img[1].to<int>();
  tensor<int> b=img[2].to<int>();
  printf("%s\n",r.toInfo());
  
  tensor<double> itp(1,2);
  
  tensor<double> lsx=tensor<double>::index(1,500);
  tensor<double> lsy=tensor<double>::index(1,400);
  lsx*=1/(double)(lsx.numel-1);
  lsy*=1/(double)(lsy.numel-1);
  //lsx-=0.5;
  //lsy-=0.5;
  //printf("%s\n",lsx.toString());
  //printf("%s\n",lsy.toString());
  tensor<uchar> hir(3,3,lsx.numel,lsy.numel);
  for (int c0=0;c0<lsx.numel;++c0)
  for (int c1=0;c1<lsy.numel;++c1) {
    itp(0)=lsx(c0);
    itp(1)=lsy(c1);
    hir(0,c0,c1)=r.interp(itp);
    hir(1,c0,c1)=g.interp(itp);
    //hir(2,c0,c1)=random<double>::uniform()*256;  
    hir(2,c0,c1)=b.interp(itp);
  }
  format::bmp("interpdone.bmp",hir);
  
  //double sdf=148.9;
  //uchar chr=sdf;
  //printf("%i\n",chr);
  
  tensor<int> tst(2,3,4);
  tst(0,0)=0;
  tst(1,0)=10;
  tst(0,1)= 1;
  tst(1,1)=101;
  tst(0,2)= 20;
  tst(1,2)=1021;
  itp(0)=0.5;
  itp(1)=1-MATH_1_3;
  printf("%s\n",tst.toString("%4i "));
  
  printf("%s\n",itp.toString());
  
  printf("\nINTERP %i\n",tst.interp(itp));
}

void tensor_appendop() {
  tensor<double> que(1,0);
  printf("%s\n",que.toString());
  for (int c0=0;c0<10;++c0)
    que<<=(double)c0+random<double>::uniform();
  printf("%s\n",que.toString());
}

tensor<double> glD(const tensor<double>& myTensor) { // ok
  assert(myTensor.numel==3||myTensor.numel==4);
  return tensor<double>(2,4,4).as(
    0.,myTensor.data[2],-myTensor.data[1],0.,
    -myTensor.data[2],0.,myTensor.data[0],0.,
    myTensor.data[1],-myTensor.data[0],0.,0.,
    0.,0.,0.,0.);
}

tensor<double> glExp(tensor<double> ome) {
  tensor<double> mat=glD(ome);
  double fac=ome|ome; // norm squared
  double sna,cna;
  sna=+1.0+fac*(
      -1/6.0+fac*(
      +1/120.0+fac*(
      -1/5040.0+fac*(
      +1/362880.0+fac*(
      -1/39916800.0+fac*(
      +1/6227020800.0+fac*(
      -1/1307674368000.0
      )))))));
  cna=+1/2.0+fac*(
      -1/24.0+fac*(
      +1/720.0+fac*(
      -1/40320.0+fac*(
      +1/3628800.0+fac*(
      -1/479001600.0+fac*(
      +1/87178291200.0+fac*(
      -1/20922789888000.0
      )))))));
  //return tensor<double>::eye(4)+mat*sna+mat*mat*cna;
  return tensor<double>::eye(4)+mat*sna+mat*mat*cna;
}

void tensor_exp_toTest() {
  tensor<double> rnd=tensor<double>::rand(1,4);
  
}

void tensor_select_toTest() {
  tensor<double> asd=tensor<double>::rand(1,20);
  for (int c0=0;c0<asd.numel;++c0) {
    double sel=asd.select(c0);
    printf(" %2i  %+f \n",c0,sel);
  }
  printf(" median %+f \n",asd.median());
}

tensor<double> newton_myF(tensor<double> x) {
  tensor<double> ret(1,1);
  ret(0)=1-cos(4*x(0))+x(0)*x(0)*0.01;
  return ret;
}

double newton_myf(tensor<double> x) {
  return newton_myF(x)|newton_myF(x)*0.5;
}

tensor<double> newton_mydF(tensor<double> x) {
  tensor<double> ret(2,x.numel,x.numel);
  ret(0,0)=x(0)*0.02+4*sin(4*x(0));
  return ret;
}

tensor<double> newton_myg(tensor<double> x) {
  tensor<double> ret(1,1);
  ret(0)=(1 + x(0)*x(0)*0.01 - cos(4*x(0)))*(x(0)*0.02 + 4*sin(4*x(0)));
  return ret;
}


void tensor_newton_toTest() {
  tensor<double> x(1,1); x(0)=0.4; // ok: startvalue
  while (x.newton(newton_myf,newton_myg(x),newton_mydF(x).inverse()*newton_myF(x)*-1,1))
    printf("x=%s  f=%f\n",x.toString(),newton_myf(x));  
  
  if (newton_myf(x)<0.001)
    printf("] yey!!! found zero!\n");
  else
    printf("] no ... fail !\n");
}

void control_toTest_V() {
  control<double> ctr;
  ctr.CN=2.3;
  ctr.MU=2;
  ctr.vn=.2;
  ctr.vd=-1.8;
  linked<tensor<double>> list;
  for (int c0=0;c0<10000;++c0) {    
    ctr.V();    
    ctr.integrate();
    list.add(ctr.toTensor());
  }
  tensor<double> arr=list.toTensor(5);
  arr.store("test_V.arr");
}

void control_toTest_VBND() {
  control<double> ctr;
  ctr.CN=2;
  ctr.MU=10;
  ctr.vn=-0.4;
  ctr.vd=0.5;
  ctr.pn=3;
  linked<tensor<double>> list;
  for (int c0=0;c0<10000;++c0) {    
    ctr.V(-2,2);
    ctr.integrate();
    list.add(ctr.toTensor());
  }
  tensor<double> arr=list.toTensor(5);
  arr.store("test_VBND.arr");
}

void control_toTest_VP() {
  control<double> ctr;
  ctr.CN=2;
  ctr.MU=10;
  ctr.vn=.2;
  //ctr.vd=-1.8;
  ctr.pn=-1.2;
  ctr.pd= 3;
  linked<tensor<double>> list;
  for (int c0=0;c0<10000;++c0) {    
    ctr.VP();    
    ctr.integrate();
    list.add(ctr.toTensor());
  }
  tensor<double> arr=list.toTensor(5);
  arr.store("test_VP.arr");
}


void control_toTestEXACT_V() {
  control<double> ctr;
  ctr.CN=2.3;
  ctr.MU=2;
  ctr.vn=.2;
  ctr.vd=-1.8;
  linked<tensor<double>> list;
  for (int c0=0;c0<10000;++c0) {    
    ctr.EXACT_V();    
    ctr.EXACT_integrate();
    list.add(ctr.toTensor());
  }
  tensor<double> arr=list.toTensor(5);
  arr.store("test_exactV.arr");
}

void control_toTestEXACT_VBND() {
  control<double> ctr;
  ctr.CN=2;
  ctr.MU=1;
  ctr.vn=-0.4;
  ctr.vd=0.5;
  ctr.pn=3;
  linked<tensor<double>> list;
  for (int c0=0;c0<10000;++c0) {    
    ctr.EXACT_V(-2,2);
    ctr.EXACT_integrate();
    list.add(ctr.toTensor());
  }
  tensor<double> arr=list.toTensor(5);
  arr.store("test_exactVBND.arr");
}

void control_toTestEXACT_VP() {
  control<double> ctr;
  ctr.CN=2;
  ctr.MU=1;
  ctr.vn=.2;
  ctr.vd=-1.3;
  ctr.pn=-1.2;
  ctr.pd= 3;
  linked<tensor<double>> list;
  for (int c0=0;c0<10000;++c0) {    
    ctr.EXACT_VP();    
    ctr.EXACT_integrate();
    list.add(ctr.toTensor());
  }
  tensor<double> arr=list.toTensor(5);
  arr.store("test_exactVP.arr");
}

void control_toTest() {
  control<double> ctr;
  
  switch (2) {     
    case 0:
      ctr.vn= -4.9116e-003;
      ctr.pn= -1.2431e+000;
      ctr.vd=-9.00e-001;
      ctr.pd= -2.6455e+000;
      for (int c0=0;c0<30000;++c0) {
        ctr.VP();
        ctr.integrate();
        printf("%s\n",ctr.toString());
        if (abs(ctr.vn-ctr.vd)<0.05||abs(ctr.pn-ctr.pd)<0.05) {
          printf("success\n");
          break;
        }
      }
      break;
    case 1: {
      int count=0;
      for (int c1=0;c1<1000;++c1) {
        linked<tensor<double>> list;
        ctr.pd=random<double>::normal()*2;
        ctr.vd=math::Clip(-0.9,random<double>::normal(),0.9);
        ctr.VP();
        double old_cn=ctr.cn;
        int bang=0;
        for (int c0=0;c0<10000;++c0) {
          list.add(ctr.toTensor());
          ctr.VP();
          if (old_cn*ctr.cn<0) 
            if (!bang) {
              bang=1;
              old_cn=ctr.cn;
            } else {
              ctr.cn=old_cn;
            }        
          ctr.integrate();
          //printf("%s\n",ctr.toString());
          if (abs(ctr.vn-ctr.vd)<0.05&&abs(ctr.pn-ctr.pd)<0.05)
            break;
        }
        if ((abs(ctr.vn-ctr.vd)<0.05&&abs(ctr.pn-ctr.pd)<0.05)&& 0.01<random<double>::uniform()) //  
          ;
        else {
          if (count<5) {
            tensor<double> arr=list.toTensor(5);
            char str[80];
            sprintf(str,"control%i.mat",++count);
            arr.store(str);
          }
          printf("%s\n",ctr.toString());
        }
      }
      break;
    }    
    case 2: {
      int count=0;
      for (int c1=0;c1<3000;++c1) {
        linked<tensor<double>> list;
        ctr.pd=random<double>::normal()*2;
        ctr.vd=math::Clip(-0.9,random<double>::normal(),0.9);
        //ctr.VP();
        //double old_cn=ctr.cn;
        //int bang=0;
        for (int c0=0;c0<10000;++c0) {
          list.add(ctr.toTensor());
          ctr.EXACT_VP();
          //if (old_cn*ctr.cn<0) 
          //  if (!bang) {
          //    bang=1;
          //    old_cn=ctr.cn;
          //  } else {
          //    ctr.cn=old_cn;
          //  }        
          ctr.EXACT_integrate();
          //printf("%s\n",ctr.toString());
          if (abs(ctr.vn-ctr.vd)<0.05&&abs(ctr.pn-ctr.pd)<0.05)
            break;
        }
        if ((abs(ctr.vn-ctr.vd)<0.05&&abs(ctr.pn-ctr.pd)<0.05)) //   && 0.01<random<double>::uniform()
          ;
        else {
          if (count<5) {
            tensor<double> arr=list.toTensor(5);
            char str[80];
            sprintf(str,"control%i.mat",++count);
            arr.store(str);
          }
          printf("%s\n",ctr.toString());
        }
      }
      break;
    }    
  }
}


//----------------------------------------------------------------------

template<class type>
void tensor_polar_toTest() {
  tensor<type> smp=tensor<type>::randn(2,3,3);
  tensor<type> pol;  
  tensor<type> ret=smp.polar(pol);  
}


void covariance_toTest(void) {
  tensor<double> mat=tensor<double>::randn(2,3,7);
  tensor<double> mea;
  tensor<double> ret=mat.covariance(mea);
  
  printf("%s\n\n",mat.toString());  
  printf("%s\n\n",ret.toString());
}

void disjoint_toTest(void) {
  int c1,c2;
  tensor<uchar> tmp=format::bmp("disjoint.bmp");
  tensor<int> img(2,tmp.size[1],tmp.size[2]);
  for (c1=0;c1<img.numel;c1++)
    img.data[c1]=tmp.data[c1*tmp.size[0]];
  
  int cnt=0;
  disjoint<int>*  dis[76800];
  tensor<int>     msk(1,76800);
  for (c1=img.size[0];c1<img.numel;c1++) {
    if (img.data[c1]) {
      cnt++;
      msk.data[c1]=cnt;                    // store address instead of color! to identifz stones            
      dis[cnt]=new disjoint<int>(c1);   // MAKESET, node has pixel address as content      
      if (msk.data[c1-1])                  // check previous pixel
        dis[cnt]->join(dis[msk.data[c1-1]]);    // UNION
      if (msk.data[c1-320])                // check pixel one row above
        dis[cnt]->join(dis[msk.data[c1-320]]);  // UNION
    }
  }
  int pins[100];  
  int pin,ref=0,num=0;                        // number of stones
  for (c1=1;c1<=cnt;c1++) {                   // loop over all yellow pixels / nodes of union-data struct collecting unique representatives
    if (pin=dis[c1]->find()->data) {
      pins[num++]=pin;
      dis[c1]->find()->data=0;
    }
  }
  printf("num %i\n",num);
  for (c1=1;c1<=cnt;c1++) {
    if (1<dis[c1]->rank)
      printf("%i ",dis[c1]->rank);
    delete dis[c1];                   // delete disjoint set union data struct
  }
}

template<class type>
void tensor_median_toTest() {
  tensor<type> smp=tensor<type>::randn(1,7);
  int res;
  type asd=smp.median(res);
  printf("%s\n",smp.toString());
  printf("%s\n",smp.sort().toString());
  printf("%f\n",asd);
  printf("%i\n",res);
}

double hash_toTensorSort_toTest_func(double key) {
  return key;
}

double tensorSortFunc(tensor<double> key) {
  assert(key.numel);
  return key.data[key.numel-1];
}


void hash_toTensorSort_toTest() {
  hash<double,int> hsh(10);
  hsh.add(-12.  ,1);
  hsh.add(-987.3,2);
  hsh.add(-2.   ,3);
  hsh.add(-87.3 ,4);
  hsh.add(-1.3  ,5);
  hsh.add(97.3 ,6);
  hsh.add(2.4  ,7);
  hsh.add(8.3  ,8);
  
  printf("%s\n",hsh.toInfo());
  
  printf("%s\n",hsh.toTensor().toString());
  tensor<double> key;
  printf("%s\n",hsh.toTensor(hash_toTensorSort_toTest_func,key).toString());
  printf("%s\n",key.toString());
  int tok;
  for (int c1=0;c1<key.numel;c1++) {
    hsh.get(key(c1),tok);
    printf("%10.2e \t %i \n",key(c1),tok);
  }
}

void hashTensor_toTensorSort_toTest() {
  hash<tensor<double>,tensor<char>> hsh(10);
  
  //tensor<double> arr=tensor<double>::vector(3).as(1.2,1.2,4.4);
  hsh.add(tensor<double>::vector(3).as(1.2,1.2,4.4),tensor<char>::string("cahr3") );
  hsh.add(tensor<double>::vector(3).as(1.2,0.2,0.4),tensor<char>::string("ad3") );
  hsh.add(tensor<double>::vector(3).as(1.2,1.2,-.3),tensor<char>::string("123ddd") );
  hsh.add(tensor<double>::vector(3).as(1.2,0.2,3.4),tensor<char>::string("d99sda") );
  
  tensor<tensor<double>> key;
  tensor<tensor<char>> str=hsh.toTensor(tensorSortFunc,key);
  
  //printf("%s\n",hsh.toTensor(hash_toTensorSort_toTest_func,key).toString());
//  printf("%s\n",key.toString());
  tensor<char> tok;
  for (int c1=0;c1<key.numel;c1++) {
    hsh.get(key(c1),tok);
    printf("%s\n",key(c1).toString());    
    printf("%s\n",str(c1).toString());    
    printf("%s\n",tok.toString());    
    //printf("%10.2e \t %i \n",key(c1),tok);
  }
}

void tensor_norm2_toTest() {
  tensor<double> vec=tensor<double>::randn(2,3,5);
  printf("%s\n\n",vec.toString());
  printf("%f\n\n",vec.norm2());
  printf("%f\n\n",vec.transpose().norm2());
  tensor<int> in0(1,2); in0.as(0,2);
  tensor<int> in1(1,3); in1.as(1,2,1);
  printf("%s\n\n",vec(in0,in1).toString());
}

void format_txt_toTest() {
  tensor<char> myf=format::txt("ed.txt");
  printf("%sEOF\n",myf.toInfo());
  printf("%sEOF\n",myf.toString());
}

void format_bmp_toTest() {
  tensor<uchar> myf(3,4,10,10);
  myf(0,2,4)=200;
  myf(1,2,4)=200;
  myf(2,2,4)=200;
  myf(3,2,4)=200;
  format::bmp("out/asd.bmp",myf);
}

void blue_toTest(void) {
  time soc=time::now();
  blue<double>   asd(tensor<int>::vector(2).as(100,100),0.025);
  tensor<double> ins(1,2);
  asd.complete(20);
  asd.empty();
  printf("%s\n",asd.toInfo());
  asd.complete(20);
  printf("time == %f \n\n",(time::now()-soc).ms());
  tensor<double> arr=asd.point.toTensor(2);
  arr.store("jans.bin");
  printf("%s\n",asd.toInfo());
}

void hash_flush(const hash<int,float>& myHash,char *name) {
  FILE *fid=fopen(name,"w");
  if (fid) {
    bucket<int,float> buck;
    for (int c0=0;c0<myHash.table.numel;++c0)
      while (myHash.table.data[c0]->token(buck))
        fprintf(fid,"%i\t%i\t%f\n",c0,buck.key,buck.data);
    fclose(fid);
  }
}

void hash_flush(const hash<tensor<int>,tensor<float>>& myHash,char *name) {
  FILE *fid=fopen(name,"w");
  if (fid) {
    bucket<tensor<int>,tensor<float>> buck;
    for (int c0=0;c0<myHash.table.numel;++c0)
      while (myHash.table.data[c0]->token(buck)) {
        printf("%s\n",buck.data.toString());
        fprintf(fid,"[%i]\t\t",c0);
        fprintf(fid,"%s\t\t",buck.key.toString());
        fprintf(fid,"%s\n",buck.data.toString());
      }
    fclose(fid);
  }
}

void tensor_eig_toTest(void) {
  printf("                       EIG test\n");
  tensor<double> mat=tensor<double>::randn(2,7,7);
  mat=mat+mat.transpose();
  printf("%s\n\n",mat.toString());
  tensor<double> ret;
  tensor<double> dia=mat.eig(ret);
  printf("%s\n\n",dia.toString());
}

void tensor_interp1_toTest(void) {
  tensor<double> asd=tensor<double>::index(1,3);
  tensor<double> itp(1,1);
  asd+=3;
  printf("%s\n\n",asd.toString());  
  for (double c1=-0.4;c1<1.3;c1+=0.03) {
    itp.data[0]=c1;
    printf("%8.3f %f\n",itp.data[0],asd.interp( itp));
  }
}

void tensor_interp2_toTest(void) {
  tensor<double> asd=tensor<double>::index(2,3,2);
  asd+=2;
  tensor<double> itp(1,2);
  printf("%s\n\n",asd.toString());  
  for (double c0=-0.1;c0<=1.1;c0+=0.03) {
    for (double c1=-0.1;c1<=1.1;c1+=0.09) {
      itp.as(c0,c1);
      printf("%5.2f",asd.interp(itp));
    }
    printf("\n");
  } 
  
}

void tensor_interp3_toTest(void) {
  tensor<double> asd=tensor<double>::index(3,3,3,3);
  asd+=3;
  tensor<double> itp(1,3);
  printf("%s\n\n",asd.toString());  
  itp.as(0.25,0.15,0.1);
  itp.as(0.5,0.5,0.51);
  printf("%f\n",asd.interp(itp));
}

void tensor_mean_var_toTest(void) {
  tensor<double> asd=(tensor<double>::rand(2,4,3)*10).to<double>();
  printf("%s\n\n",asd.toString());
  printf("%s\n\n",asd.mean().toString());
  printf("%s\n\n",asd.variance().toString());
}

void tensor_toTest(void) {
  tensor<double> asd=tensor<double>::rand(2,3,3);
  tensor<double> ret;
  tensor<double> qrt=asd.schur(ret);
  printf("%s\n---\n",asd.toString());
  printf("%s\n---\n",ret.chop().toString());
  printf("%s\n---\n",(qrt*ret*qrt.transpose()).toString());
  printf("double\n%s\n", asd.toString());
  printf("int   \n%s\n",(asd*0x7fffffff).to<int>().toString("%08x "));
  printf("--------------------------\n");
  tensor<double> mat=tensor<double>::rand(4,4,3,3,2);
  printf("%s\n\n",mat.toString());
  printf("%s\n\n",mat.cumsize().toString());
  printf("%s\n\n",mat.mean().toString());
  printf("%s\n\n",((mat[0]+mat[1])*0.5).toString());
  assert(mat.mean()==(mat[0]+mat[1])*0.5);
  printf("--------------------------\n");

  //asd.to<double>();
}

template<class type>
void tensor_trace_toTest(void) {
  tensor<int> rnd=(tensor<double>::rand(2,4,4)*4).to<int>();
  printf("%s \n\n",rnd.toString());
  printf("%s \n\n",(rnd==1).toString());
  printf("%s \n\n",(rnd==1).find().toString());
  printf("%s \n",rnd.trace(0,1).toString());
  printf("%i \n",rnd.trace(0,1).data[0]);
}

template<class type>
void tensor_qr_toTest(void) {
  tensor<type> rnd=tensor<type>::randn(2,7,7);
  printf("%s \n\n",rnd.toString());
  tensor<type> ret;
  tensor<type> qor=rnd.qr(ret);
  printf("%s \n\n",qor.toString());
  printf("%s \n\n",ret.toString());
  tensor<type> dif=qor*rnd-ret;
  printf("%s \n\n",dif.toString());
  dif=qor*qor.transpose();
  printf("%s \n\n",dif.toString());
  if (rnd.size[0]==rnd.size[1]) {
    type prd=1;
    for (int c0=0;c0<rnd.size[0];++c0)
      prd*=ret(c0,c0);
    printf("det = %10.2e  %10.2e\n",rnd.det(),prd);
  }
  assert((int)0==0.0);
  assert(qor*qor.transpose()==tensor<type>::eye(qor.size[0])); // qor is orthogonal
  assert(ret==qor*rnd); // R=Q*A <=> Q^t*R=A
  for (int c1=0;c1<rnd.size[1];++c1)
    for (int c0=c1+1;c0<rnd.size[0];++c0)
      assert(abs(ret(c0,c1))<TENSOR_QR); // values below diagonal are null
}

template<class type>
void tensor_svd_sparse_toTest(void) {
  tensor<tensor<type>*> blk(2,2,2);
  //blk(0,0)=new tensor<type>::rand(2,3,3);
  blk(0,0)=new tensor<type>(tensor<type>::rand(2,3,3));
  blk(1,0)=new tensor<type>(2,3,3);
  blk(0,1)=new tensor<type>(2,3,3);
  blk(1,1)=new tensor<type>(2,3,3);
  tensor<type> MAT(blk);
  tensor<type> w,v;  
  tensor<type> d=MAT.svd(w,v);
  printf("%s\n",d.toString());
  printf("%s\n",v.toString());
  printf("%s\n",w.toString());
  printf("------------------\n");
  printf("%s\n",(d.transpose()*tensor<type>::diag(w)*v.transpose()).toString());
  printf("------------------\n");
  printf("%s\n",MAT.toString());
  printf("------------------\n");
}

template<class type>
void tensor_block_toTest(void) {
  tensor<tensor<type>*> bla(2,4,3);
  for (int c0=0;c0<bla.numel;++c0)
    bla.data[c0]=new tensor<type>(2,1,2);
  tensor<type> BLA(bla);
}

template<class type>
void tensor_string_toTest(void) {
  tensor<type> asd=tensor<type>::index(2,2,3)+32;
  printf("%s\n",asd.toString());
  tensor<char> txt=tensor<type>::string((type)123.2);
  printf("-----> %s\n",txt.toString());
}

void tensor_svd_toTest(void) {
  tensor<double> svdt=tensor<double>::randn(2,11,5);
  for (int c1=0;c1<svdt.size[0];c1++)
    svdt(c1,3)=0;
  printf("%s\n\n",svdt.toString());
  tensor<double> uor;
  tensor<double> dia;
  tensor<double> ort;
  tensor<double>::valid=0;
  dia=svdt.svd(uor,ort);
  printf("%s\n\n",dia.toInfo());
  printf("valid=%i\n",tensor<double>::valid);
  printf("%s\n\n",(uor*uor.transpose()).toString());
  printf("%s\n\n",(ort.transpose()*ort).toString());
  printf("%s\n\n",(ort*ort.transpose()).toString());
  dia=tensor<double>::diag(dia);
  printf("%s\n\n",dia.toString());
  printf("difference=\n%s\n\n",(uor.transpose()*dia*ort.transpose()-svdt).toString());
  //printf("%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
  
  tensor<int> asd=tensor<int>::matrix(12,12).as(
    -2  ,   0  ,   0  ,   0  ,   0   ,  0   , -1  ,   0  ,   0  ,   0  ,   0  ,   0  ,
     0  ,  -6  ,   0  ,   0  ,  -2   ,  0   ,  0  ,  -1  ,   0  ,   0  ,   0  ,   0  ,
     0  ,   0  ,  -6  ,   0  ,   0   , -2   ,  0  ,   0  ,  -1  ,   0  ,   0  ,   0  ,
     0  ,   0  ,   0  ,   0  ,   0   ,  0   ,  0  ,   0  ,   0  ,   0  ,   0  ,   0  ,
     0  ,  -2  ,   0  ,   0  ,  -2   ,  0   ,  0  ,   0  ,   0  ,   1  ,   0  ,   0  ,
     0  ,   0  ,  -2  ,   0  ,   0   , -2   ,  0  ,   0  ,   0  ,   0  ,   0  ,   0  ,
    -1  ,   0  ,   0  ,   0  ,   0   ,  0   , -2  ,   0  ,   0  ,   0  ,   0  ,   0  ,
     0  ,  -1  ,   0  ,   0  ,   0   ,  0   ,  0  ,  -6  ,   0  ,  -2  ,   0  ,   0  ,
     0  ,   0  ,  -1  ,   0  ,   0   ,  0   ,  0  ,   0  ,  -6  ,   0  ,   0  ,   0  ,
     0  ,   0  ,   0  ,   0  ,   1   ,  0   ,  0  ,  -2  ,   0  ,  -2  ,   0  ,   0  ,
     0  ,   0  ,   0  ,   0  ,   0   ,  0   ,  0  ,   0  ,   0  ,   0  ,   0  ,   0  ,
     0  ,   0  ,   0  ,   0  ,   0   ,  0   ,  0  ,   0  ,   0  ,   0  ,   0  ,  -2  );
     
  tensor<double> tst=asd.to<double>();
  tensor<double> ret,dgh;
  tst.svd(ret,dgh);

  
}

template<class type>
void tensor_lu_toTest() {
  tensor<int> ind;
  type sgn;
  tensor<type> mlu=tensor<type>::randn(2,4,4);
  tensor<type> rhs=tensor<type>::randn(1,4);
  tensor<type> ret=mlu.lu(ind,sgn);
  tensor<type> sol=ret.lu(ind,rhs);
  
  printf("%s\n\n",(mlu*sol).toString());
  printf("%s\n\n",rhs.toString());
  
  printf("%s\n\n",sol.toString());
  printf("%s\n\n",mlu.solve(rhs).toString());
}

template<class type>
void tensor_inverse_toTest() {
  tensor<type> rnd=tensor<type>::rand(2,7,5);
  tensor<type> spl=tensor<type>::eye(5);
  spl(4,3)=spl(3,4)=spl(1,2)=spl(2,1)=1;
  rnd=rnd*spl;
  //rnd=rnd.transpose();
  printf("rank = %i\n\n",rnd.rank());
  printf("FOR MATLAB\n%s\n---------\n",rnd.toString());
  tensor<type> ide=rnd.inverse();
  printf("OUR PINVERSE:\n%s\n--------------\n",ide.toString());
  //printf("%s\n",ide.toString());
  //tensor<type> fgh=tensor<type>::randn(2,7,7);
  //tensor<type> dif=fgh.solve(tensor<type>::eye(7))-fgh.inverse();
  //printf("inv=pinv :\n%s\n",dif.toString());
}

void tensor_rank_toTest(void) {
  tensor<double> svdt=tensor<double>::randn(2,11,11);
  for (int c1=0;c1<svdt.size[0];c1++) {    
    //svdt(c1,1)=TENSOR_NO_DIVISION;
    //svdt(c1,3)=0;
  }
  printf("rank = %i\n\n",svdt.transpose().rank());
}

void tensor_sort_inv_test() {
  tensor<int> rnd=(tensor<float>::rand(1,15)*100).to<int>();
  tensor<int> ind;
  tensor<int> srt(ind);
  //tensor<int> srt=rnd.sort(ind);
  printf("%s\n",srt.toString());
}

void tensor_svd_toTest2(void) {
  //tensor<double> svdt=tensor<double>::randn(2,6,6);
  tensor<double> svdt(2,5,5);//=tensor<double>::randn(2,6,6);
  //for (int c1=0;c1<svdt.size[0];c1++) {
  //  svdt(c1,3)=0;
  //  svdt(c1,4)=0;
  //  svdt(c1,5)=0;
  //}
  printf("%s\n\n",svdt.toString());
  tensor<double> uor;
  tensor<double> dia;
  tensor<double> ort;
  tensor<double>::valid=0;
  printf("valid=%i\n",tensor<double>::valid);
  uor=svdt.svd(dia,ort);
  printf("valid=%i\n",tensor<double>::valid);
  printf("%s\n\n",uor.toString());
  printf("%s\n\n",(uor.transpose()*uor).toString());
  printf("%s\n\n",(ort.transpose()*ort).toString());
  printf("%s\n\n",(ort*ort.transpose()).toString());
  dia=tensor<double>::diag(dia);
  printf("%s\n\n",dia.toString());
  printf("difference=\n%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
  //printf("%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
}

void sha_toTest() {
  sha mySha;
  char* stra=new char[16];
  stra[ 0]=0x0;
  stra[ 1]=0x1;
  stra[ 2]=0x2;
  stra[ 3]=0x3;
  stra[ 4]=0x4;
  stra[ 5]=0x5;
  stra[ 6]=0x6;
  stra[ 7]=0x7;
  stra[ 8]=0x8;
  stra[ 9]=0x9;
  stra[10]=0x10;
  stra[11]=0x11;
  stra[12]=0x12;
  stra[13]=0x13;
  stra[14]=0x14;
  stra[15]=0x15;
  mySha.add(stra,16);
  mySha.add(stra,16);
  mySha.add(stra,16);
  mySha.add(stra,16);
  //mySha.add(stra,16);
  //mySha.add(stra,10);
  //mySha.add(stra,10);
  //mySha.add(stra,10);
  printf("%s\n",mySha.toInfo());
  printf("%s\n",mySha.toString());
  //tensor<uint> asd2=mySha.flush();
  tensor<int> asd=mySha.flush();
  printf("\n%s\n3fc78b03 7eb94ea8 d0cc96b2 c1a172b7 3df20a50\n",asd.toString("%08x "));
  //printf("%s\n",mySha.toInfo());
  //printf("%s\n",mySha.toString());
  //printf("%08x\n",asd.sum());
  //mySha.add(stra,3);
  //printf("%s\n",mySha.toInfo());
  //asd=mySha.flush();  
  //printf("%s\n",asd.toString("%08x "));
  //printf("%08x\n",asd.sum());
  //tensor<int> dist(1,13);  
  //for (int c1=0;c1<10000;c1++) {
  //  tensor<double> some=tensor<double>::rand(2,10,4);
  //  //printf("%s\n",some.toInfo());
  //  //printf("%s\n",some.toString());
  //  //printf("%08x\n",sha_hash(some));
  //  //printf("%08x\n",sha::hash<double>(some));
  //  //printf("%i\n",sha::hash<double>(some));
  //  dist.data[sha::hash<double>(some)%13]++;
  //}
  //for (int c1=0;c1<13;c1++) 
  //  printf("%3i %i\n",c1,dist.data[c1]);
}

void random_toTest(void) {
  //random<double>::randomize();
  printf("%s\n",random<double>::toInfo());
  double d1=random<double>::uniform();
  printf("%s\n",random<double>::toInfo());
  double n1,n2;
  for (int c1=0;c1<10;c1++) {
    n1=random<double>::normal();
    printf("%f\n",n1);
  }
}

//----------------------------------------------------------------------

void complex_toTest() {
  complex<double> a(-2,3);
  complex<double> b(5,9);
  printf("%s\n",(a/b).toInfo());
  printf("%s\n",(a*b).toInfo());
}

void node_toTest(void) {
  node<int> myNodeA(0,3);
  node<int> myNodeB(&myNodeA,2);
  printf("%s\n",myNodeA.toInfo());
  printf("%s\n",myNodeB.toInfo());
}

//----------------------------------------------------------------------

void linked_toTest(void) {
  int myInt;
  linked<int> myLinked;
  printf("%s\n",myLinked.toInfo());
  myLinked.add(0);
  printf("%s <- 1 added\n",myLinked.toInfo());
  while (myLinked.token(myInt))
   myLinked.remove();
  printf("%s <- 1 removed \n",myLinked.toInfo());
  myLinked.add(1);
  myLinked.add(2);
  myLinked.add(3);
  myLinked.add(4);
  myLinked.add(5);
  myLinked.add(6);
  printf("%s\n",myLinked.toInfo());
  
  while (myLinked.token(myInt)) {
    printf("token %i\n",myInt);
    if (myInt==3||myInt==1) {
      myLinked.remove();
      myLinked.add(0);      
      myLinked.pre(-1);
    }
  }
  printf("%s\n",myLinked.toInfo());
  while (myLinked.token(myInt))
    printf("token %i\n",myInt);
  //myLinked.token_reset();
  printf("----\n");
  printf("%s\n",myLinked.toInfo());  
  while (myLinked.token(myInt)) {
    printf("token %i\n",myInt);
  }
  tensor<int> myTensor=myLinked.toTensor();
  printf("tensor %s\n",myTensor.toString());
  printf("%s\n",myLinked.toInfo());
  while (myLinked.token(myInt))
    myLinked.remove();    
  printf("%s <- after while remove \n",myLinked.toInfo());  
  myLinked.empty();
  printf("%s\n",myLinked.toInfo());  
  printf("-------\n");
  linked<int>* myStar=new linked<int>();
  printf("%s\n",myStar->toInfo());
  myStar->add(0);
  myStar->add(1);
  printf("%s\n",myStar->toInfo());
  delete myStar;
  //printf("%s\n",myStar->toInfo());  
}

int hash_toTest_int(int myInt) {
  return myInt%HASH_TOTEST;
}

void hash_toTest_int(void) {
  hash<int,float> myHash(HASH_TOTEST,set<int>::map);  
  float myFloat=0.23f;
  myHash.add(3,myFloat);
  myFloat=99;
  myHash.put(3,myFloat);
  //myHash.remove(3);
  //myHash.add(3,myFloat);
  printf("float %8.3f\n",myFloat);
  myFloat=-10;
  myHash.add(3+5,myFloat);
  myFloat=.3333;
  myHash.add(3+25,myFloat);
  bucket<int,float> buck;
  for (int c0=0;c0<myHash.table.numel;++c0)
    while (myHash.table(c0)->token(buck))
      printf("%i key=%4i data=%8.3f\n",c0,buck.key,buck.data);
  printf("%s\n",myHash.toInfo());
  int status;
  status=myHash.get(3+2,myFloat);
  printf("get 5 %i %6.3f\n",status,myFloat);
  status=myHash.get(3+5,myFloat);
  printf("get 8 %i %6.3f\n",status,myFloat);
  hash_flush(myHash,"myhash.txt");  
}

void hash_toTest_void(void) {
  hash<int,void*> myHash(HASH_TOTEST,set<int>::map);  
  void* null=(void*)9999;
  myHash.add(3,null);
  int status;
  status=myHash.get(3,null);
  printf("get 3 %i %08x\n",status,null);
}


int hash_toTest_tensor(tensor<int> myTensor) {
  return myTensor.sum()%HASH_TOTEST;
}

void hash_toTest_tensor(void) {
  hash<tensor<int>,tensor<float>> myHash(HASH_TOTEST,set<tensor<int>>::map);
  tensor<int> key(1,3);
  key.as(2,3,4);
  tensor<float> color(1,3);
  color(0)=0.222;
  color(1)=0.111;
  color(2)=0.333;
  myHash.add(key,color);
  key(1)++;
  myHash.add(key,color);
  
  printf("%s\n",myHash.toInfo());
  hash_flush(myHash,"myhash_tensor.txt");
}

void set_toTest(void) {
  set<int> myS(10,set<int>::map);
  myS.toggle(3);
  myS.toggle(10);
  myS.toggle(13);
  printf("%s\n",myS.toInfo());
  myS.toggle(3);
  myS.toggle(10);
  myS.toggle(13);  
  printf("%s\n",myS.toInfo());
  
  set<tensor<int>> myV(11,set<tensor<int>>::map);
  tensor<int> inp(1,3);
  inp.data[2]=1;
  myV.toggle(inp);
  printf("%s\n",myV.toInfo());
  inp.data[1]=2;
  myV.toggle(inp);
  printf("%s\n",myV.toInfo());  
}

void hash_toTest_newest(void) {
  hash<int,int> myV(1,set<int>::map);
  myV.add(123,3);
  myV.add(12,144);
  printf("%s\n",myV.toString());
}

void tensor_toTest_char(void) {
  tensor<char> st1=tensor<char>::string("get/meshgl/jaja");
  tensor<char> st2=tensor<char>::string("get/meshgl/jaj");
  printf("%i\n",st1==st2);
  printf("%s\n",st1.toString());  
}

void random_M_toTest() {
  random<double> myr;
  double nxt;
  for (int c1=0;c1<100;++c1) {
    for (int c0=0;c0<1e7;++c0) 
      if (myr.uniform()==0.0)
        printf("%i null!\n",c0);
    printf("%03i\n",c1);
  }
}

void tensorChar_toTest_sort() {
  tensor<tensor<char>> str=tensor<tensor<char>>(1,5);
  str(0)=tensor<char>::string("get/meshgl/jaja");
  str(1)=tensor<char>::string("123");
  str(2)=tensor<char>::string(" kajhsg");
  str(3)=tensor<char>::string("ahsg");
  str(4)=tensor<char>::string("ah33sg");
  
  tensor<int> ind=tensor<int>::index(1,str.numel);
  int flag=1;
  int buf;
  while (flag) {
    flag=0;
    for (int c1=1;c1<str.numel;++c1)
      if (0<strncmp(str(ind(c1-1)).data,str(ind(c1)).data,min(str(ind(c1-1)).numel,str(ind(c1)).numel))) {
        buf      =ind(c1-1);
        ind(c1-1)=ind(c1);
        ind(c1  )=buf;
        flag=1;
      }
  }
  
  for (int c1=0;c1<ind.numel;++c1)
    printf("%i  %s\n",ind(c1),str(ind(c1)).toString());
    
    
  printf("\n");
  
  ind=0;
  ind=tensor_char_order(str);  
  for (int c1=0;c1<ind.numel;++c1)
    printf("%i  %s\n",ind(c1),str(ind(c1)).toString());
}
