// code by jph

// since there is no copyconstructor, or operator= defined
//  => always use  const rapid&  when passing!

#include "../_extern/RAPID.H"

// static char RAPID_BUFFER[128];

// TODO only simply point(s) for connected components!!!

class rapid_touch {
public:
  int            num; // number of contacts
  //tensor<int>    par; // 2 x num with indices to triangles
  tensor<double> pnt;
  tensor<double> nrm;
  double         str; // 
};

class rapid {
public:
  rapid(meshgl* msh);
  ~rapid(void);

  tensor<double> pos;
  RAPID_model* model;
  tensor<double> are; // volume of triangle to compute weighted averages
  tensor<double> V4D; // 3 x T+2Q avg points of shapes
  tensor<double> N4D; // 3 x T+2Q avg normal or shapes

  static int collide(rapid& rap0,rapid& rap1);
  static int collide(rapid& rap0,rapid& rap1,rapid_touch& tch);
};

rapid::rapid(meshgl* msh) { 
  pos=tensor<double>::eye(4);
  model=new RAPID_model();
  model->BeginModel();
  tensor<double> v0(1,4); v0(3)=1;
  tensor<double> v1(1,4); v1(3)=1;
  tensor<double> v2(1,4); v2(3)=1;
  int ofs=msh->offsetV3F();
  tensor<double> crs;
  tensor<int> idx=tensor<int>::index(1,msh->ind[MESH_QUAD].size[1]);
  tensor<int> tri=msh->ind[MESH_TRIA]&msh->ind[MESH_QUAD](tensor<int>::vector(3).as(3,0,1),idx)&msh->ind[MESH_QUAD](tensor<int>::vector(3).as(1,2,3),idx);
  are=tensor<double>(1,tri.size[1]);
  V4D=tensor<double>(2,4,tri.size[1]);
  N4D=tensor<double>(2,4,tri.size[1]);
  for (int c0=0;c0<tri.size[1];++c0) {
    v0.in(0,msh->arr[tri(0,c0)].to<double>()^ofs,3);
    v1.in(0,msh->arr[tri(1,c0)].to<double>()^ofs,3);
    v2.in(0,msh->arr[tri(2,c0)].to<double>()^ofs,3);
    model->AddTri(v0.data,v1.data,v2.data,c0);
    crs=(v1-v0).x(v2-v0);
    are(c0)=crs.norm()*0.5;                   
    V4D.in(c0*4,(v0+v1+v2)*MATH_1_3);
    N4D.in(c0*4,crs.normal());
  }
  model->EndModel();
}

inline rapid::~rapid(void) {
  delete model;
}

int rapid::collide(rapid& rap0,rapid& rap1) {
  assert(rap0.model!=rap1.model); // TODO why is this a criteria
  int c0,id0,id1;
  double rot0[3][3];
  double rot1[3][3];
  rot0[0][0]=rap0.pos.data[ 0]; rot0[0][1]=rap0.pos.data[ 4]; rot0[0][2]=rap0.pos.data[ 8];
  rot0[1][0]=rap0.pos.data[ 1]; rot0[1][1]=rap0.pos.data[ 5]; rot0[1][2]=rap0.pos.data[ 9];
  rot0[2][0]=rap0.pos.data[ 2]; rot0[2][1]=rap0.pos.data[ 6]; rot0[2][2]=rap0.pos.data[10];
  rot1[0][0]=rap1.pos.data[ 0]; rot1[0][1]=rap1.pos.data[ 4]; rot1[0][2]=rap1.pos.data[ 8];
  rot1[1][0]=rap1.pos.data[ 1]; rot1[1][1]=rap1.pos.data[ 5]; rot1[1][2]=rap1.pos.data[ 9];
  rot1[2][0]=rap1.pos.data[ 2]; rot1[2][1]=rap1.pos.data[ 6]; rot1[2][2]=rap1.pos.data[10];
  RAPID_Collide(rot0,rap0.pos^12,rap0.model, rot1,rap1.pos^12,rap1.model, RAPID_FIRST_CONTACT);
  return RAPID_num_contacts; // returns 0 or 1
}

int rapid::collide(rapid& rap0,rapid& rap1,rapid_touch &tch) {
  assert(rap0.model!=rap1.model); // TODO why is this a criteria
  int c0,id0,id1;
  double rot0[3][3];
  double rot1[3][3];
  rot0[0][0]=rap0.pos.data[ 0]; rot0[0][1]=rap0.pos.data[ 4]; rot0[0][2]=rap0.pos.data[ 8];
  rot0[1][0]=rap0.pos.data[ 1]; rot0[1][1]=rap0.pos.data[ 5]; rot0[1][2]=rap0.pos.data[ 9];
  rot0[2][0]=rap0.pos.data[ 2]; rot0[2][1]=rap0.pos.data[ 6]; rot0[2][2]=rap0.pos.data[10];
  rot1[0][0]=rap1.pos.data[ 0]; rot1[0][1]=rap1.pos.data[ 4]; rot1[0][2]=rap1.pos.data[ 8];
  rot1[1][0]=rap1.pos.data[ 1]; rot1[1][1]=rap1.pos.data[ 5]; rot1[1][2]=rap1.pos.data[ 9];
  rot1[2][0]=rap1.pos.data[ 2]; rot1[2][1]=rap1.pos.data[ 6]; rot1[2][2]=rap1.pos.data[10];
  RAPID_Collide(rot0,rap0.pos^12,rap0.model, rot1,rap1.pos^12,rap1.model, RAPID_ALL_CONTACTS);
  tch.num=RAPID_num_contacts;
  //tch.par=tensor<int>(2,2,RAPID_num_contacts);
  //int c1=-1;
  //for (c0=0;c0<RAPID_num_contacts;++c0) {
  //  tch.par.data[++c1]=RAPID_contact[c0].id1;
  //  tch.par.data[++c1]=RAPID_contact[c0].id2;
  //}
  tensor<double>::valid=1;
  tch.str=0;
  if (tch.num) {
    double sca;
    tensor<double> nrm0;
    tensor<double> nrm1;
    tensor<double> pnt(1,4);
    tensor<double> nrm(1,4);
    for (c0=0;c0<RAPID_num_contacts;++c0) {
      id0=RAPID_contact[c0].id1;
      id1=RAPID_contact[c0].id2;
      if (rap0.are.data[id0]<rap1.are.data[id1]) { // smaller triangle contributes to position
        assert(id0<rap0.V4D.size[1]);
        pnt+=rap0.pos*rap0.V4D[id0];
      } else {
        assert(id1<rap1.V4D.size[1]);
        pnt+=rap1.pos*rap1.V4D[id1];
      }
      nrm0=rap0.pos*rap0.N4D[id0];
      nrm1=rap1.pos*rap1.N4D[id1];
      sca=-min(0,nrm0|nrm1)*min(rap0.are.data[id0],rap1.are.data[id1]);
      nrm+=(nrm1-nrm0)*sca;
      tch.str+=sca;
    }
    tch.pnt=pnt*(1/pnt(3));
    tch.nrm=nrm.normal();
  }
  return tch.num&&tensor<double>::valid;
}

